@@ -11,7 +11,7 @@
 # See the file "license.terms" for information on usage and redistribution
 # of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #
-# RCS: @(#) $Id: httpdthread.tcl,v 1.11.6.2 2002/10/29 06:12:22 welch Exp $
+# RCS: @(#) $Id: httpdthread.tcl,v 1.19 2004/04/23 02:11:22 coldstore Exp $
 
 # Note about per-thread vs. per-application.  Essentially all
 # the "package require" commands are needed in all the threads,
@@ -33,7 +33,22 @@ package require httpd          	;# Protocol stack
 package require httpd::version	;# Version number
 package require httpd::url	;# URL dispatching
 package require httpd::mtype	;# Mime types
-Mtype_ReadTypes 		[file join $Config(lib) mime.types]
+
+# Search for mime.types either right in Config(lib), or down
+# one level in the installed tclhttpd subdirectory
+
+foreach path [list \
+    [file join $Config(lib) mime.types] \
+    [glob -nocomplain [file join $Config(lib) tclhttpd* mime.types]] \
+    ] {
+  if {[llength $path] > 0} {
+    set path [lindex $path 0]
+  }
+  if {[file exists $path]} {
+    Mtype_ReadTypes $path
+    break
+  }
+}
 package require httpd::counter	;# Statistics
 Counter_Init $Config(secs)
 package require httpd::utils	;# handy stuff like "lassign"
@@ -41,6 +56,7 @@ package require httpd::utils	;# handy stuff like "lassign"
 package require httpd::redirect	;# URL redirection
 package require httpd::auth	;# Basic authentication
 package require httpd::log	;# Standard logging
+package require httpd::digest	;# Digest authentication
 
 if {$Config(threads) > 0} {
     package require Thread		;# C extension
@@ -65,9 +81,6 @@ package require httpd::doc
 # your web-visible file structure.
 
 Doc_Root			$Config(docRoot)
-if {[file isdirectory [file join $Config(docRoot) libtml]]} {
-    Doc_TemplateLibrary [file join $Config(docRoot) libtml]
-}
 
 # Merge in a second file system into the URL tree.
 
@@ -76,15 +89,15 @@ if {[file isdirectory $htdocs_2]} {
     Doc_AddRoot /addroot	$htdocs_2
 }
 
-# Doc_TemplateInterp determines which interpreter to use when
+# Template_Interp determines which interpreter to use when
 # interpreting templates.
 
-Doc_TemplateInterp		{}
+Template_Interp		{}
 
 # Doc_IndexFile defines the name of the default index file
 # in each directory.  Its value is a glob pattern.
 
-Doc_IndexFile			index.{tml,html,shtml,thtml,htm,subst}
+DirList_IndexFile			index.{tml,html,shtml,thtml,htm,subst}
 
 # Doc_PublicHtml turns on the mapping from ~user to the
 # specified directory under their home directory.
@@ -96,7 +109,7 @@ Doc_PublicHtml			public_html
 # up-to-date.  If the .tml (or its dependent files) are newer than
 # the HTML file, the HTML file is regenerated from the template.
 
-Doc_CheckTemplates		1
+Template_Check		1
 
 # Doc_ErrorPage registers a template to be used when a page raises an
 # uncaught Tcl error.  This is a crude template that is simply passed through
@@ -113,11 +126,22 @@ Doc_NotFoundPage		/notfound.html
 # Doc_Webmaster returns the value last passed into it.
 # Designed to be used in page templates where contact email is needed.
 
-Doc_Webmaster			$Config(webmaster)
+Httpd_Webmaster			$Config(webmaster)
 
+# uncomment the following and commment out the package requires
+# if you want to leave out dirlist support
+#proc Dir_ListingIsHidden {} {
+#    return 1
+#}
 package require httpd::dirlist		;# Directory listings
 package require httpd::include		;# Server side includes
 
+# uncomment this and comment the package requires
+# if you want to leave out cgi support
+#proc Cgi_Domain {virtual directory sock suffix} {
+#	Doc_NotFound $sock
+#	return
+#}
 package require httpd::cgi		;# Standard CGI
 Cgi_Directory			/cgi-bin
 
@@ -140,6 +164,8 @@ Debug_Url			/debug
 package require httpd::redirect	;# Url redirection tables
 Redirect_Init			/redirect
 
+package require httpd::doctools ;# doctool type conversions
+
 if {[catch {
     Auth_InitCrypt			;# Probe for crypt module
 } err]} {
@@ -163,7 +189,7 @@ if {[info exist Config(library)] && [string length $Config(library)]} {
 	    Stderr "Loading code from $Config(library)"
 	}
 	foreach f [lsort -dictionary [glob -nocomplain [file join $Config(library) *.tcl]]] {
-            if {[string compare $f "pkgIndex.tcl"] == 0} {
+            if {[string compare [file tail $f] "pkgIndex.tcl"] == 0} {
               continue
             } elseif {[catch {source $f} err]} {
 		Stderr "$f: $err"
