@@ -10,11 +10,12 @@
 # See the file "license.terms" for information on usage and redistribution
 # of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 # 
-# RCS: @(#) $Id: pop3.tcl,v 1.17 2002/01/18 20:51:16 andreas_kupries Exp $
+# RCS: @(#) $Id: s.pop3.tcl 1.28 04/01/21 21:33:53-08:00 tcl@bluepeak.home $
 
 package require Tcl 8.2
 package require cmdline
-package provide pop3 1.5.1
+package require log
+package provide pop3 1.6.1
 
 namespace eval ::pop3 {
 
@@ -28,8 +29,8 @@ namespace eval ::pop3 {
     # The value of "msex" influences how the translation for the
     # channel is set and is determined by the contents of the received
     # greeting. The value of "retr_mode" is initially "retr" and
-    # completely determined by the first call to [retr]. For "list" the
-    # system will use LIST before RETR to retrieve the message size.
+    # completely determined by the first call to [retrieve]. For "list"
+    # the system will use LIST before RETR to retrieve the message size.
 
     # The state can be influenced by options given to "open".
 
@@ -38,6 +39,21 @@ namespace eval ::pop3 {
 
 }
 
+# ::pop3::config --
+#
+#	Retrieve configuration of pop3 connection
+#
+# Arguments:
+#	chan      The channel, returned by ::pop3::open
+#
+# Results:
+#	A serialized array.
+
+proc ::pop3::config {chan} {
+    variable state
+    return  $state($chan)
+}
+
 # ::pop3::close --
 #
 #	Close the connection to the POP3 server.
@@ -78,8 +94,10 @@ proc ::pop3::close {chan} {
 #       May throw errors from the server.
 
 proc ::pop3::delete {chan start {end -1}} {
-    
-    set count [lindex [::pop3::status $chan] 0]
+
+    variable state
+    array set  cstate $state($chan)
+    set count $cstate(limit)
     set last 0
     catch {set last [::pop3::last $chan]}
 
@@ -208,7 +226,9 @@ proc ::pop3::list {chan {msg ""}} {
 
 proc ::pop3::open {args} {
     variable state
-    array set cstate {msex 0 retr_mode retr}
+    array set cstate {msex 0 retr_mode retr limit {}}
+
+    log::log debug "pop3::open | [join $args]"
 
     while {[set err [cmdline::getopt args {msex.arg retr-mode.arg} opt arg]]} {
 	if {$err < 0} {
@@ -248,11 +268,15 @@ proc ::pop3::open {args} {
 	set port 110
     }
 
+    log::log debug "pop3::open | protocol, connect to $host $port"
+
     # Argument processing is finally complete, now open the channel
 
     set chan [socket $host $port]
     fconfigure $chan -buffering none
 
+    log::log debug "pop3::open | connect on $chan"
+
     if {$cstate(msex)} {
 	# We are talking to MS Exchange. Work around its quirks.
 	fconfigure $chan -translation binary
@@ -260,7 +284,10 @@ proc ::pop3::open {args} {
 	fconfigure $chan -translation {binary crlf}
     }
 
+    log::log debug "pop3::open | wait for greeting"
+
     if {[catch {::pop3::send $chan {}} errorStr]} {
+	::close $chan
 	error "POP3 CONNECT ERROR: $errorStr"
     }
 
@@ -272,16 +299,31 @@ proc ::pop3::open {args} {
 	fconfigure $chan -translation binary
     }
 
+    log::log debug "pop3::open | authenticate $user (*password not shown*)"
+
     if {[catch {
-	    ::pop3::send $chan "user $user"
-	    ::pop3::send $chan "pass $password"
-        } errorStr]} {
+	::pop3::send $chan "user $user"
+	::pop3::send $chan "pass $password"
+    } errorStr]} {
+	::close $chan
 	error "POP3 LOGIN ERROR: $errorStr"
     }
 
+    # [ 833486 ] Can't delete messages one at a time ...
+    # Remember the number of messages in the maildrop at the beginning
+    # of the session. This gives us the highest possible number for
+    # message ids later. Note that this number must not be affected
+    # when deleting mails later. While the number of messages drops
+    # down the limit for the message id's stays the same. The messages
+    # are not renumbered before the session actually closed.
+
+    set cstate(limit) [lindex [::pop3::status $chan] 0]
+
     # Remember the state.
 
     set state($chan) [array get cstate]
+
+    log::log debug "pop3::open | ok ($chan)"
     return $chan
 }
 
@@ -295,7 +337,7 @@ proc ::pop3::open {args} {
 #                 May be "next" (the next message after the last
 #                 one seen, see ::pop3::last), "start" (aka 1),
 #                 "end" (the last message in the spool, for 
-#                 retriving only the last message).
+#                 retrieving only the last message).
 #       end       (optional, defaults to -1) The last message
 #                 to retrieve in the range. May be "last"
 #                 (the last message viewed), "end" (the last
@@ -311,7 +353,7 @@ proc ::pop3::retrieve {chan start {end -1}} {
     variable state
     array set cstate $state($chan)
     
-    set count [lindex [::pop3::status $chan] 0]
+    set count $cstate(limit)
     set last 0
     catch {set last [::pop3::last $chan]}
 
@@ -349,31 +391,39 @@ proc ::pop3::retrieve {chan start {end -1}} {
     
     set result {}
 
+    ::log::log debug "pop3 $chan retrieve $start -- $end"
+
     for {set index $start} {$index <= $end} {incr index} {
 	switch -exact -- $cstate(retr_mode) {
 	    retr {
 		set sizeStr [::pop3::send $chan "RETR $index"]
 
-		if {[scan $sizeStr {%d %s} size dummy] < 0} {
+		::log::log debug "pop3 $chan retrieve ($sizeStr)"
+
+		if {[scan $sizeStr {%d %s} size dummy] < 1} {
 		    # The server did not deliver the size information.
 		    # Switch our mode to "list" and use the slow
 		    # method this time. The next call will use LIST before
 		    # RETR to get the size information. If even that fails
 		    # the system will fall back to slow mode all the time.
 
+		    ::log::log debug "pop3 $chan retrieve - no size information, go slow"
+
 		    set cstate(retr_mode) list
 		    set state($chan) [array get cstate]
 
 		    # Retrieve in slow motion.
 		    set msgBuffer [RetrSlow $chan]
 		} else {
+		    ::log::log debug "pop3 $chan retrieve - size information present, use fast mode"
+
 		    set msgBuffer [RetrFast $chan $size]
 		}
 	    }
 	    list {
 		set sizeStr [::pop3::send $chan "LIST $index"]
 
-		if {[scan $sizeStr {%d %d %s} dummy size dummy] < 0} {
+		if {[scan $sizeStr {%d %d %s} dummy size dummy] < 2} {
 		    # Not even LIST generates the necessary size information.
 		    # Switch to full slow mode and don't bother anymore.
 
@@ -389,7 +439,6 @@ proc ::pop3::retrieve {chan start {end -1}} {
 		    # through LIST
 
 		    ::pop3::send $chan "RETR $index"
-
 		    set msgBuffer [RetrFast $chan $size]
 		}
 	    }
@@ -419,17 +468,101 @@ proc ::pop3::retrieve {chan start {end -1}} {
 proc ::pop3::RetrFast {chan size} {
     set msgBuffer [read $chan $size]
 
-    # We might have read not enough because of .-stuffed lines.
-    # Read the possible remainder in line by line fashion!
-    #		    
-    # get the terminating "."
-    # sometimes the gets returns nothing, 
-    # need to get the real terminating "."
-
-    while {[set line [gets $chan]] != ".\r"} {
-	append msgBuffer $line
+    foreach line [split $msgBuffer \n] {
+	::log::log debug "pop3 $chan fast <$line>"
     }
 
+    # There is a small discrepance in counting octets we have to be
+    # aware of. 'size' is #octets before transmission, i.e. can be
+    # with one eol character, CR or LF. The channel system in binary
+    # mode counts every character, and the protocol specified CRLF as
+    # eol, so for every line in the message we read that many
+    # characters _less_. Another factor which can cause a miscount is
+    # the ".-stuffing performed by the sender. I.e. what we got now is
+    # not necessarily the complete message. We have to perform slow
+    # reads to get the remainder of the message. This has another
+    # complication. We cannot simply check for a line containing the
+    # terminating signature, simply because the point where the
+    # message was broken in two might just be in between the dots of a
+    # "\r\n..\r\n" sequence. We have to make sure that we do not
+    # misinterpret the second part of this sequence as terminator.
+    # Another possibility: "\r\n.\r\n" is broken just after the dot.
+    # Then we have to ensure to not to miss the terminator entirely.
+
+    # Sometimes the gets returns nothing, need to get the real
+    # terminating "."                                    / "
+
+    if {[string equal [string range $msgBuffer end-3 end] "\n.\r\n"]} {
+	# Complete terminator found. Remove it from the message buffer.
+
+	::log::log debug "pop3 $chan /5__"
+	set msgBuffer [string range $msgBuffer 0 end-3]
+
+    } elseif {[string equal [string range $msgBuffer end-2 end] "\n.\r"]} {
+	# Complete terminator found. Remove it from the message buffer.
+	# Also perform an empty read to remove the missing '\n' from
+	# the channel. If we don't do this all following commands will
+	# run into off-by-one (character) problems.
+
+	::log::log debug "pop3 $chan /4__"
+	set msgBuffer [string range $msgBuffer 0 end-2]
+	while {[read $chan 1] != "\n"} {}
+
+    } elseif {[string equal [string range $msgBuffer end-1 end] "\n."]} {
+	# \n. at the end of the fast buffer.
+	# Can be	\n.\r\n	 = Terminator
+	# or		\n..\r\n = dot-stuffed single .
+
+	log::log debug "pop3 $chan /check for cut .. or terminator sequence"
+
+	# Idle until non-empty line encountered.
+	while {[set line [gets $chan]] == ""} {}
+	if {"$line" == "\r"} {
+	    # Terminator already found. Note that we have to
+	    # remove the partial terminator sequence from the
+	    # message buffer.
+	    ::log::log debug "pop3 $chan /3__ <$line>"
+	    set msgBuffer [string range $msgBuffer 0 end-1]
+	} else {
+	    # Append line and look for the real terminator
+	    append msgBuffer $line
+	    ::log::log debug "pop3 $chan ____ <$line>"
+	    while {[set line [gets $chan]] != ".\r"} {
+		::log::log debug "pop3 $chan ____ <$line>"
+		append msgBuffer $line
+	    }
+	    ::log::log debug "pop3 $chan /2__ <$line>"
+	}
+    } elseif {[string equal [string index $msgBuffer end] \n]} {
+	# Line terminator (\n) found. The remainder of the mail has to
+	# consist of true lines we can read directly.
+
+	while {![string equal [set line [gets $chan]] ".\r"]} {
+	    ::log::log debug "pop3 $chan ____ <$line>"
+	    append msgBuffer $line
+	}
+	::log::log debug "pop3 $chan /1__ <$line>"
+    } else {
+	# Incomplete line at the end of the buffer. We complete it in
+	# a single read, and then handle the remainder like the case
+	# before, where we had a complete line at the end of the
+	# buffer.
+
+	set line [gets $chan]
+	::log::log debug "pop3 $chan /1a_ <$line>"
+	append msgBuffer $line
+
+	::log::log debug "pop3 $chan /1b_"
+
+	while {![string equal [set line [gets $chan]] ".\r"]} {
+	    ::log::log debug "pop3 $chan ____ <$line>"
+	    append msgBuffer $line
+	}
+	::log::log debug "pop3 $chan /1c_ <$line>"
+    }
+
+    ::log::log debug "pop3 $chan done"
+
     # Map both cr+lf and cr to lf to simulate auto EOL translation, then
     # unstuff .-stuffed lines.
 
@@ -448,11 +581,13 @@ proc ::pop3::RetrFast {chan size} {
 #	The text of the retrieved message.
 
 proc ::pop3::RetrSlow {chan} {
+
     set msgBuffer ""
 	
     while {1} {
 	set line [string trimright [gets $chan] \r]
-	    
+	::log::log debug "pop3 $chan slow $line"
+
 	# End of the message is a line with just "."
 	if {$line == "."} {
 	    break
@@ -483,10 +618,12 @@ proc ::pop3::send {chan cmdstring} {
    global PopErrorNm PopErrorStr debug
 
    if {$cmdstring != {}} {
-      puts $chan $cmdstring
+       ::log::log debug "pop3 $chan >>> $cmdstring"       
+       puts $chan $cmdstring
    }
    
    set popRet [string trim [gets $chan]]
+   ::log::log debug "pop3 $chan <<< $popRet"
 
    if {[string first "+OK" $popRet] == -1} {
        error [string range $popRet 4 end]
