@@ -11,15 +11,15 @@
 # more details.
 # -------------------------------------------------------------------------
 
+package require Tcl 8.3;                # tcl minimum version
 package require log;                    # tcllib
 package require mime;                   # tcllib
 
-namespace eval smtpd {
-    variable rcsid {$Id: smtpd.tcl,v 1.2 2001/12/10 21:13:20 patthoyts Exp $}
-    variable version 1.0
+namespace eval ::smtpd {
+    variable rcsid {$Id: s.smtpd.tcl 1.11 03/05/01 00:17:41-00:00 andreas_kupries $}
+    variable version 1.2.1
     variable stopped
 
-    package provide smtpd $version
     namespace export start stop
 
     variable commands {EHLO HELO MAIL RCPT DATA RSET NOOP QUIT}
@@ -28,19 +28,29 @@ namespace eval smtpd {
     variable options
     if {! [info exists options]} {
         array set options {
+            serveraddr         {}
+            deliverMIME        {}
             deliver            {}
             validate_host      {}
             validate_sender    {}
             validate_recipient {}
         }
     }
+
+    variable extensions
+    if {! [info exists extensions]} {
+        array set extensions {
+            8BITMIME {}
+            SIZE     0
+        }
+    }
 }
 
 # -------------------------------------------------------------------------
 # Description:
 #   Obtain configuration options for the server.
 #
-proc smtpd::cget {option} {
+proc ::smtpd::cget {option} {
     variable options
     set optname [string trimleft $option -]
     if { [info exists options($optname)] } {
@@ -59,7 +69,7 @@ proc smtpd::cget {option} {
 #   resources once the deliver proc has completed.
 #   An example might be to exec procmail to deliver the mail to users.
 #
-proc smtpd::configure {args} {
+proc ::smtpd::configure {args} {
     variable options
 
     if {[llength $args] == 0} {
@@ -72,13 +82,15 @@ proc smtpd::configure {args} {
 
     foreach {opt value} $args {
         switch -- $opt {
+            -deliverMIME        {set options(deliverMIME) $value}
             -deliver            {set options(deliver) $value}
             -validate_host      {set options(validate_host) $value}
             -validate_sender    {set options(validate_sender) $value}
             -validate_recipient {set options(validate_recipient) $value}
             default {
                 error "unknown option: \"$opt\": must be one of \
-                       -deliver, -validate_host, -validate_recipient \
+                       -deliverMIME, -deliver,\
+                       -validate_host, -validate_recipient \
                        or -validate_sender"
             }
         }
@@ -90,7 +102,7 @@ proc smtpd::configure {args} {
 # Description:
 #   Start the server on the given interface and port.
 #
-proc smtpd::start {{myaddr {}} {port 25}} {
+proc ::smtpd::start {{myaddr {}} {port 25}} {
     variable options
     variable stopped
     
@@ -99,8 +111,14 @@ proc smtpd::start {{myaddr {}} {port 25}} {
     }
 
     if {$myaddr != {}} {
+        set options(serveraddr) $myaddr
         set myaddr "-myaddr $myaddr"
+    } else {
+        if {$options(serveraddr) == {}} {
+            set options(serveraddr) [info hostname]
+        }
     }
+
     set options(socket) [eval socket \
                              -server [namespace current]::accept $myaddr $port]
     set stopped 0
@@ -112,7 +130,7 @@ proc smtpd::start {{myaddr {}} {port 25}} {
 # Description:
 #  Stop a running server. Do nothing if the server isn't running.
 #
-proc smtpd::stop {} {
+proc ::smtpd::stop {} {
     variable options
     variable stopped
     if {[info exists options(socket)]} {
@@ -128,17 +146,17 @@ proc smtpd::stop {} {
 #   Accept a new connection and setup a fileevent handler to process the new
 #   session. Performs a host id validation step before allowing access.
 #
-proc smtpd::accept {channel client_addr client_port} {
+proc ::smtpd::accept {channel client_addr client_port} {
     variable options
     variable version
+    upvar [namespace current]::state_$channel State
 
     # init state array
-    # FRINK: nocheck
-    catch {unset [namespace current]::state_$channel}
-    state $channel access allowed
-    state $channel indata 0
-    state $channel client_addr $client_addr
-    state $channel client_port $client_port
+    catch {unset State}
+    initializeState $channel
+    set State(access) allowed
+    set State(client_addr) $client_addr
+    set State(client_port) $client_port
     set accepted true
 
     # configure the data channel
@@ -149,8 +167,8 @@ proc smtpd::accept {channel client_addr client_port} {
     if {[cget -validate_host] != {}} {
         if {[catch {eval [cget -validate_host] $client_addr} msg] } {
             log::log notice "access denied for $client_addr:$client_port: $msg"
-            puts $channel "550 Access denied: $msg"
-            state $channel access denied
+            Puts $channel "550 Access denied: $msg"
+            set State(access) denied
             set accepted false
         }
     }
@@ -158,19 +176,32 @@ proc smtpd::accept {channel client_addr client_port} {
     if {$accepted} {
         # Accept the connection
         log::log notice "connect from $client_addr:$client_port on $channel"
-        puts $channel "220 [info hostname] tcllib smtpd $version; [timestamp]"
+        Puts $channel "220 $options(serveraddr) tcllib smtpd $version; [timestamp]"
     }
     
     return
 }
 
+# -------------------------------------------------------------------------
+# Description:
+#   Initialize the channel state array. Called by accept and RSET.
+#
+proc ::smtpd::initializeState {channel} {
+    upvar [namespace current]::state_$channel State
+    set State(indata) 0
+    set State(to) {}
+    set State(from) {}
+    set State(data) {}
+    set State(options) {}
+}
+
 # -------------------------------------------------------------------------
 # Description:
 #   Access the state of a connected session using the channel name as part
 #   of the state array name. Called with no value, it returns the current
 #   value of the item (or {} if not defined).
 #
-proc smtpd::state {channel args} {
+proc ::smtpd::state {channel args} {
     if {[llength $args] == 0} {
         return [array get [namespace current]::state_$channel]
     }
@@ -185,7 +216,7 @@ proc smtpd::state {channel args} {
         }
         return $r
     }
-    
+
     foreach {name value} $args {
         # FRINK: nocheck
         set [namespace current]::[subst state_$channel]($name) $value
@@ -193,35 +224,74 @@ proc smtpd::state {channel args} {
     return {}
 }
 
+# -------------------------------------------------------------------------
+# Description:
+#   Safe puts.
+#   If the client closes the channel, then puts will throw an error. Lets
+#   terminate the session if this occurs.
+proc ::smtpd::Puts {channel args} {
+    if {[catch {uplevel puts $channel $args} msg]} {
+        log::log error $msg
+        catch {
+            close $channel
+            # FRINK: nocheck
+            unset -- [namespace current]::state_$channel
+        }
+    }
+    return $msg
+}
+
 # -------------------------------------------------------------------------
 # Description:
 #   Perform the chat with a connected client. This procedure accepts input on
 #   the connected socket and executes commands according to the state of the
 #   session.
 #
-proc smtpd::service {channel} {
+proc ::smtpd::service {channel} {
     variable commands
     variable options
-    
+    upvar [namespace current]::state_$channel State
+
     if {[eof $channel]} {
         close $channel
         return
     }
-    
-    gets $channel cmdline
+
+    if {[catch {gets $channel cmdline} msg]} {
+        close $channel
+        log::log error $msg
+        return
+    }
+
+    if { $cmdline == "" && [eof $channel] } {
+        log::log warning "client has closed the channel"
+        return
+    }
+
     log::log debug "received: $cmdline"
 
     # If we are handling a DATA section, keep looking for the end of data.
-    if {[state $channel indata] } {
+    if {$State(indata)} {
         if {$cmdline == "."} {
-            state $channel indata 0
+            set State(indata) 0
             fconfigure $channel -translation crlf
-            puts $channel "250 [state $channel id]\
-                            Message accepted for delivery"
-            deliver $channel
+            if {[catch {deliver $channel} err]} {
+                # permit delivery handler to return SMTP errors in errorCode
+                if {[regexp {\d{3}} $::errorCode]} {
+                    Puts $channel "$::errorCode $err"
+                } else {
+                    Puts $channel "554 Transaction failed: $err"
+                }
+            } else {
+                Puts $channel "250 [state $channel id]\
+                        Message accepted for delivery"
+            }
         } else {
-            # FRINK: nocheck
-            lappend [namespace current]::[subst state_$channel](data) $cmdline
+            # RFC 2821 section 4.5.2: Transparency
+            if {[string match {..*} $cmdline]} {
+                set cmdline [string range $cmdline 1 end]
+            }
+            lappend State(data) $cmdline
         }
         return
     }
@@ -230,19 +300,19 @@ proc smtpd::service {channel} {
     set cmd [string toupper [lindex [split $cmdline] 0]]
     if {[lsearch $commands $cmd] != -1} {
         if {[info proc $cmd] == {}} {
-            puts $channel "500 $cmd not implemented"
+            Puts $channel "500 $cmd not implemented"
         } else {
             # If access denied then client can only issue QUIT.
-            if {[state $channel access] == "denied" && $cmd != "QUIT" } {
-                puts $channel "503 bad sequence of commands"
+            if {$State(access) == "denied" && $cmd != "QUIT" } {
+                Puts $channel "503 bad sequence of commands"
             } else {
                 set r [eval $cmd $channel [list $cmdline]]
             }
         }
     } else {
-        puts $channel "500 Invalid command"
+        Puts $channel "500 Invalid command"
     }
-    
+
     return
 }
 
@@ -250,7 +320,7 @@ proc smtpd::service {channel} {
 # Description:
 #  Generate a random ASCII character for use in mail identifiers.
 #
-proc smtpd::uidchar {} {
+proc ::smtpd::uidchar {} {
     set c .
     while {! [string is alnum $c]} {
         set n [expr {int(rand() * 74 + 48)}]
@@ -262,7 +332,7 @@ proc smtpd::uidchar {} {
 # Description:
 #  Generate a unique random identifier using only ASCII alphanumeric chars.
 #
-proc smtpd::uid {} {
+proc ::smtpd::uid {} {
     set r {}
     for {set cn 0} {$cn < 12} {incr cn} {
         append r [uidchar]
@@ -274,7 +344,7 @@ proc smtpd::uid {} {
 # Description:
 #   Calculate the local offset from GMT in hours for use in the timestamp
 #
-proc smtpd::gmtoffset {} {
+proc ::smtpd::gmtoffset {} {
     set now [clock seconds]
     set lh [string trimleft [clock format $now -format "%H" -gmt false] 0]
     set zh [string trimleft [clock format $now -format "%H" -gmt true] 0]
@@ -296,7 +366,7 @@ proc smtpd::gmtoffset {} {
 #   Generate a standard SMTP compliant timestamp. That is a local time but with
 #   the timezone represented as an offset.
 #
-proc smtpd::timestamp {} {
+proc ::smtpd::timestamp {} {
     set ts [clock format [clock seconds] \
                 -format "%a, %d %b %Y %H:%M:%S" -gmt false]
     append ts " " [gmtoffset]
@@ -307,20 +377,58 @@ proc smtpd::timestamp {} {
 # Description:
 #   Get the servers ip address (from http://purl.org/mini/tcl/526.html)
 #
-proc smtpd::server_ip {} {
+proc ::smtpd::server_ip {} {
     set me [socket -server xxx -myaddr [info hostname] 0]
     set ip [lindex [fconfigure $me -sockname] 0]
     close $me
     return $ip
 }
 
+# -------------------------------------------------------------------------
+# Description:
+#   deliver is called once a mail transaction is completed and there is
+#   no deliver procedure defined
+#   The configured -deliverMIME procedure is called with a MIME token.
+#   If no such callback is defined then try the -deliver option and use
+#   the old API.
+#
+proc ::smtpd::deliver {channel} {
+    set deliverMIME [cget deliverMIME]
+    if { $deliverMIME != {} \
+            && [state $channel from] != {} \
+            && [state $channel to] != {} \
+            && [state $channel data] != {} } {
+        
+        # create a MIME token from the mail message.        
+        set tok [mime::initialize -string \
+                [join [state $channel data] "\n"]]
+#        mime::setheader $tok "From" [state $channel from]
+#        foreach recipient [state $channel to] {
+#            mime::setheader $tok "To" $recipient -mode append
+#        }
+        
+        # catch and rethrow any errors.
+        set err [catch {$deliverMIME $tok} msg]
+        mime::finalize $tok -subordinates all
+        if {$err} {
+            log::log debug "error in deliver: $msg"
+            return -code error -errorcode $::errorCode \
+                    -errorinfo $::errorInfo $msg
+        }        
+        
+    } else {
+        # Try the old interface
+        deliver_old $channel
+    }
+}
+
 # -------------------------------------------------------------------------
 # Description:
 #   Deliver is called once a mail transaction is completed (defined as the
 #   completion of a DATA command). The configured -deliver procedure is called
 #   with the sender, list of recipients and the text of the mail.
 #
-proc smtpd::deliver {channel} {
+proc ::smtpd::deliver_old {channel} {
     set deliver [cget deliver]
     if { $deliver != {} \
              && [state $channel from] != {} \
@@ -330,10 +438,22 @@ proc smtpd::deliver {channel} {
                         [state $channel to] \
                         [state $channel data]} msg]} {
             log::log debug "error in deliver: $msg"
+            return -code error -errorcode $::errorCode \
+                    -errorinfo $::errorInfo $msg
         }
     }
 }
 
+# -------------------------------------------------------------------------
+proc ::smtpd::split_address {address} {
+    set start [string first < $address]
+    set end [string last > $address]
+    set addr [string range $address $start $end]
+    incr end
+    set opts [string trim [string range $address $end end]]
+    return [list $addr $opts]
+}
+
 # -------------------------------------------------------------------------
 # The SMTP Commands
 # -------------------------------------------------------------------------
@@ -342,47 +462,56 @@ proc smtpd::deliver {channel} {
 # Reference:
 #   RFC2821 4.1.1.1
 #
-proc smtpd::HELO {channel line} {
+proc ::smtpd::HELO {channel line} {
+    variable options
+
     if {[state $channel domain] != {}} {
-        puts $channel "503 bad sequence of commands"
+        Puts $channel "503 bad sequence of commands"
         log::log debug "HELO received out of sequence."
         return
     }
 
     set r [regexp -nocase {^HELO\s+([-\w\.]+)\s*$} $line -> domain]
     if {$r == 0} {
-        puts $channel "501 Syntax error in parameters or arguments"
+        Puts $channel "501 Syntax error in parameters or arguments"
         log::log debug "HELO received \"$line\""
         return
     }
-    puts $channel "250-[info hostname] Hello $domain\
+    Puts $channel "250-$options(serveraddr) Hello $domain\
                      \[[state $channel client_addr]\], pleased to meet you"
-    puts $channel "250 Ready for mail."
+    Puts $channel "250 Ready for mail."
     state $channel domain $domain
     log::log debug "HELO on $channel from $domain"
     return
 }
 
+# -------------------------------------------------------------------------
 # Description:
 #   Initiate an ESMTP session
 # Reference:
 #   RFC2821 4.1.1.1
-proc smtpd::EHLO {channel line} {
+proc ::smtpd::EHLO {channel line} {
+    variable options
+    variable extensions
+
     if {[state $channel domain] != {}} {
-        puts $channel "503 bad sequence of commands"
+        Puts $channel "503 bad sequence of commands"
         log::log debug "EHLO received out of sequence."
         return
     }
 
     set r [regexp -nocase {^EHLO\s+([-\w\.]+)\s*$} $line -> domain]
     if {$r == 0} {
-        puts $channel "501 Syntax error in parameters or arguments"
+        Puts $channel "501 Syntax error in parameters or arguments"
         log::log debug "EHLO received \"$line\""
         return
     }
-    puts $channel "250-[info hostname] Hello $domain\
+    Puts $channel "250-$options(serveraddr) Hello $domain\
                      \[[state $channel client_addr]\], pleased to meet you"
-    puts $channel "250 Ready for mail."
+    foreach {extn opts} [array get extensions] {
+        Puts $channel [string trimright "250-$extn $opts"]
+    }
+    Puts $channel "250 Ready for mail."
     state $channel domain $domain
     log::log debug "EHLO on $channel from $domain"
     return
@@ -393,17 +522,24 @@ proc smtpd::EHLO {channel line} {
 # Reference:
 #   RFC2821 4.1.1.2
 #
-proc smtpd::MAIL {channel line} {
+proc ::smtpd::MAIL {channel line} {
     set r [regexp -nocase {^MAIL FROM:\s*(.*)} $line -> from]
     if {$r == 0} {
-        puts $channel "501 Syntax error in parameters or arguments"
+        Puts $channel "501 Syntax error in parameters or arguments"
         log::log debug "MAIL received \"$line\""
         return
     }
-    eval array set addr [mime::parseaddress $from]
-    if {$addr(error) != {}} {
+    if {[catch {
+        set from [split_address $from]
+        set opts [lindex $from 1]
+        set from [lindex $from 0]
+        eval array set addr [mime::parseaddress $from]
+    } msg]} {
+        set addr(error) $msg
+    }
+    if {$addr(error) != {} } {
         log::log debug "MAIL failed $addr(error)"
-        puts $channel "501 Syntax error in parameters or arguments"
+        Puts $channel "501 Syntax error in parameters or arguments"
         return
     }
 
@@ -411,7 +547,7 @@ proc smtpd::MAIL {channel line} {
         if {[catch {eval [cget -validate_sender] $addr(address)}]} {
             # this user has been denied
             log::log info "MAIL denied user $addr(address)"
-            puts $channel "553 Requested action not taken:\
+            Puts $channel "553 Requested action not taken:\
                             mailbox name not allowed"
             return
         }
@@ -419,7 +555,8 @@ proc smtpd::MAIL {channel line} {
 
     log::log debug "MAIL FROM: $addr(address)"
     state $channel from $from
-    puts $channel "250 OK"
+    state $channel options $opts
+    Puts $channel "250 OK"
     return
 }
 
@@ -434,17 +571,25 @@ proc smtpd::MAIL {channel line} {
 # Notes:
 #   The postmaster mailbox MUST be supported. (RFC2821: 4.5.1)
 #
-proc smtpd::RCPT {channel line} {
+proc ::smtpd::RCPT {channel line} {
     set r [regexp -nocase {^RCPT TO:\s*(.*)} $line -> to]
     if {$r == 0} {
-        puts $channel "501 Syntax error in parameters or arguments"
+        Puts $channel "501 Syntax error in parameters or arguments"
         log::log debug "RCPT received \"$line\""
         return
     }
-    eval array set addr [mime::parseaddress $to]
+    if {[catch {
+        set to [split_address $to]
+        set opts [lindex $to 1]
+        set to [lindex $to 0]
+        eval array set addr [mime::parseaddress $to]
+    } msg]} {
+        set addr(error) $msg
+    }
+
     if {$addr(error) != {}} {
         log::log debug "RCPT failed $addr(error)"
-        puts $channel "501 Syntax error in parameters or arguments"
+        Puts $channel "501 Syntax error in parameters or arguments"
         return
     }
 
@@ -456,7 +601,7 @@ proc smtpd::RCPT {channel line} {
             if {[catch {eval [cget -validate_recipient] $addr(address)}]} {
                 # this recipient has been denied
                 log::log info "RCPT denied mailbox $addr(address)"
-                puts $channel "553 Requested action not taken:\
+                Puts $channel "553 Requested action not taken:\
                             mailbox name not allowed"
                 return
             }
@@ -468,7 +613,7 @@ proc smtpd::RCPT {channel line} {
     catch {set recipients [state $channel to]}
     lappend recipients $to
     state $channel to $recipients
-    puts $channel "250 OK"
+    Puts $channel "250 OK"
     return
 }
 
@@ -488,20 +633,26 @@ proc smtpd::RCPT {channel line} {
 #   'service' procedure for the handling of DATA lines.
 #   We also insert trace information as per RFC2821:4.4
 #
-proc smtpd::DATA {channel line} {
+proc ::smtpd::DATA {channel line} {
     variable version
-    if { [state $channel from] != {} && [state $channel to] != {} } {
-        puts $channel "354 Enter mail, end with \".\" on a line by itself"
-        state $channel id [uid]
-        state $channel indata 1
-        eval array set sndr [mime::parseaddress [state $channel from]]
-
-        set trace "Received: from [state $channel domain] \[[state $channel client_addr]\]\n\
-              \tby [info hostname] ($version) id [state $channel id]; [timestamp]"
-        state $channel data [list $trace]
-        fconfigure $channel -translation auto
+    upvar [namespace current]::state_$channel State
+    log::log debug "DATA"
+    if { $State(from) == {}} {
+        Puts $channel "503 bad sequence: no sender specified"
+    } elseif { $State(to) == {}} {
+        Puts $channel "503 bad sequence: no recipient specified"
     } else {
-        puts $channel "503 bad sequence of commands"
+        Puts $channel "354 Enter mail, end with \".\" on a line by itself"
+        set State(id) [uid]
+        set State(indata) 1
+
+        lappend trace "Return-Path: $State(from)"
+        lappend trace "Received: from [state $channel domain]\
+                   \[[state $channel client_addr]\]"
+        lappend trace "\tby [info hostname] with tcllib smtpd ($version)\
+                   id $State(id); [timestamp]"
+        set State(data) $trace
+        fconfigure $channel -translation auto ;# naughty: RFC2821:2.3.7
     }
     return
 }
@@ -512,18 +663,13 @@ proc smtpd::DATA {channel line} {
 # Reference:
 #   RFC2821 4.1.1.5
 #
-proc smtpd::RSET {channel line} {
-    
-    if {[catch {
-        state $channel indata 0
-        state $channel from {}
-        state $channel to {}
-        state $channel data {}
-    } msg]} {
+proc ::smtpd::RSET {channel line} {
+    upvar [namespace current]::state_$channel State
+    log::log debug "RSET on $channel"
+    if {[catch {initializeState $channel} msg]} {
         log::log warning "RSET: $msg"
     }
-    puts $channel "250 OK"
-    log::log debug "RSET on $channel"
+    Puts $channel "250 OK"
     return
 }
 
@@ -533,7 +679,7 @@ proc smtpd::RSET {channel line} {
 # Reference:
 #   RFC2821 4.1.1.6
 #
-#proc smtpd::VRFY {channel line} {
+#proc ::smtpd::VRFY {channel line} {
 #    # VRFY SP String CRLF
 #}
 
@@ -543,7 +689,7 @@ proc smtpd::RSET {channel line} {
 # Reference:
 #   RFC2821 4.1.1.7
 #
-#proc smtpd::EXPN {channel line} {
+#proc ::smtpd::EXPN {channel line} {
 #    # EXPN SP String CRLF
 #}
 
@@ -553,7 +699,7 @@ proc smtpd::RSET {channel line} {
 # Reference:
 #   RFC2821 4.1.1.8
 #
-#proc smtpd::HELP {channel line} {
+#proc ::smtpd::HELP {channel line} {
 #    # HELP SP String CRLF
 #}
 
@@ -563,11 +709,11 @@ proc smtpd::RSET {channel line} {
 # Reference:
 #   RFC2821 4.1.1.9
 #
-proc smtpd::NOOP {channel line} {
+proc ::smtpd::NOOP {channel line} {
     set str {}
     regexp -nocase {^NOOP (.*)$} -> str
     log::log debug "NOOP: $str"
-    puts $channel "250 OK"
+    Puts $channel "250 OK"
     return
 }
 
@@ -580,16 +726,21 @@ proc smtpd::NOOP {channel line} {
 #   The server is only permitted to close the channel once it has received 
 #   a QUIT message.
 #
-proc smtpd::QUIT {channel line} {
+proc ::smtpd::QUIT {channel line} {
+    variable options
+    upvar [namespace current]::state_$channel State
+
     log::log debug "QUIT on $channel"
-    puts $channel "221 [info hostname] Service closing transmission channel"
+    Puts $channel "221 $options(serveraddr) Service closing transmission channel"
     close $channel
         
     # cleanup the session state array.
-    #unset [namespace current]::state_$channel
+    unset State
     return
 }
 
+package provide smtpd $smtpd::version
+
 # -------------------------------------------------------------------------
 # Local variables:
 #   mode: tcl
